from pathlib import Path

from typer import Option, Argument

from patterns.cli.services.lookup import IdLookup
from patterns.cli.services.output import sprint, abort_on_error
from patterns.cli.services.trigger import trigger_node, TypeChoices

_graph_help = "The location of the graph.yml file for the deployed graph"
_graph_version_id_help = "The id of the deployed graph version"
_environment_help = "The name of the Patterns environment that the graph is deployed to"
_organization_help = (
    "The name of the Patterns organization that the graph specified "
    "with --graph was uploaded to"
)
_node_help = "The path to the node to trigger"


def trigger(
    organization: str = Option("", "-o", "--organization", help=_organization_help),
    environment: str = Option("", "-e", "--environment", help=_environment_help),
    graph: Path = Option(None, exists=True, help=_graph_help),
    graph_version_id: str = Option("", help=_graph_version_id_help),
    type: TypeChoices = Option(TypeChoices.pubsub, hidden=True),
    node: Path = Argument(..., exists=True, help=_node_help),
):
    """Trigger a node on a deployed graph to run immediately"""
    ids = IdLookup(
        environment_name=environment,
        organization_name=organization,
        explicit_graph_path=graph,
        node_file_path=node,
        explicit_graph_version_id=graph_version_id,
        find_nearest_graph=True,
    )
    with abort_on_error("Error triggering node"):
        trigger_node(
            ids.node_id, ids.graph_version_id, ids.environment_id, execution_type=type,
        )

    sprint(f"[success]Triggered node {node}")



import json
from enum import Enum
from pathlib import Path
from typing import Iterable

import typer
from rich.table import Table
from typer import Option, Argument

from patterns.cli.services.environments import paginated_environments
from patterns.cli.services.list import (
    paginated_output_data,
    paginated_execution_events,
    paginated_graphs,
    paginated_webhook_urls,
    paginated_graph_components,
)
from patterns.cli.services.lookup import IdLookup
from patterns.cli.services.organizations import paginated_organizations
from patterns.cli.services.output import sprint, abort_on_error


class Listable(str, Enum):
    env = "environments"
    graph = "graphs"


_type_help = "The type of object to list"
_json_help = "Output the object as JSON Lines"
_organization_help = "The name of the Patterns organization to use"
_environment_help = "The name of the Patterns environment to use"

_organization_option = Option("", "--organization", "-o", help=_organization_help)
_environment_option = Option("", "--environment", "-e", help=_environment_help)

list_command = typer.Typer(name="list", help="List objects of a given type")


@list_command.command()
def graphs(
    organization: str = Option("", help=_organization_help),
    print_json: bool = Option(False, "--json", help=_json_help),
):
    """List graphs"""
    ids = IdLookup(organization_name=organization)
    with abort_on_error("Error listing graphs"):
        gs = list(paginated_graphs(ids.organization_id))
    _print_objects(gs, print_json)


@list_command.command()
def environments(
    organization: str = Option("", "--organization", "-o", help=_organization_help),
    print_json: bool = Option(False, "--json", help=_json_help),
):
    """List environments"""
    ids = IdLookup(organization_name=organization)
    with abort_on_error("Error listing environments"):
        es = list(paginated_environments(ids.organization_id))
    _print_objects(es, print_json)


@list_command.command()
def organizations(
    print_json: bool = Option(False, "--json", help=_json_help),
):
    """List organizations"""
    with abort_on_error("Error listing organizations"):
        es = list(paginated_organizations())
    _print_objects(es, print_json)


_node_help = "The path to the node to get logs for"


@list_command.command()
def logs(
    print_json: bool = Option(False, "--json", help=_json_help),
    organization: str = Option("", "-o", "--organization", help=_organization_help),
    environment: str = Option("", "-e", "--environment", help=_environment_help),
    node: Path = Argument(..., exists=True, help=_node_help),
):
    """List execution logs for a node"""
    ids = IdLookup(
        environment_name=environment,
        organization_name=organization,
        node_file_path=node,
    )

    with abort_on_error("Could not list logs"):
        events = list(
            paginated_execution_events(ids.environment_id, ids.graph_id, ids.node_id)
        )
    _print_objects(events, print_json)


_port_help = "The name of the output port"

_store_name_help = "The name of the store to get the latest output for"
_graph_help = "The location of the graph.yml file for the graph"


@list_command.command()
def output(
    print_json: bool = Option(False, "--json", help=_json_help),
    organization: str = Option("", "-o", "--organization", help=_organization_help),
    environment: str = Option("", "-e", "--environment", help=_environment_help),
    graph: Path = Option("", "-g", "--graph-path", help=_graph_help),
    store_name: str = Argument(..., exists=True, help=_store_name_help),
):
    """List data sent to an output port of a from the most recent run of a node"""
    ids = IdLookup(
        environment_name=environment,
        organization_name=organization,
        explicit_graph_path=graph,
    )

    with abort_on_error("Could not get node data"):
        data = list(
            paginated_output_data(ids.environment_id, ids.graph_id, store_name)
        )
    _print_objects(data, print_json)


_graph_help = "The location of the graph.yml file of the graph to list"


@list_command.command()
def webhooks(
    print_json: bool = Option(False, "--json", help=_json_help),
    organization: str = Option("", "-o", "--organization", help=_organization_help),
    environment: str = Option("", "-e", "--environment", help=_environment_help),
    graph: Path = Argument(None, exists=True, help=_graph_help),
):
    """List webhook urls for a graph"""
    ids = IdLookup(
        environment_name=environment,
        organization_name=organization,
        explicit_graph_path=graph,
    )
    with abort_on_error("Could not get webhook data"):
        data = list(paginated_webhook_urls(ids.environment_id, ids.graph_id))

    _print_objects(data, print_json, headers=["name", "node_id", "webhook_url"])


@list_command.command()
def components(print_json: bool = Option(False, "--json", help=_json_help)):
    """List available graph components that you can add to your graphs"""
    with abort_on_error("Could not get components"):
        data = list(paginated_graph_components())

    # flatten the latest_version dict
    for row in data:
        version = row.pop("latest_version", {})
        for k, v in version.items():
            if isinstance(v, dict) and "slug" in v:
                v = v["slug"]
            if k not in row:
                row[k] = v
        row["version_names"].remove(row["graph_version_uid"])

    _print_objects(data, print_json)


def _print_objects(objects: list, print_json: bool, headers: Iterable[str] = ()):
    if not objects:
        if not print_json:
            sprint("[info]No data found")
        return

    if print_json:
        for o in objects:
            print(json.dumps(o))
    else:
        table = Table()
        for k in headers:
            table.add_column(k)
        for k in objects[0].keys():
            if k not in headers:
                table.add_column(k)
        columns = [str(c.header) for c in table.columns]
        for o in objects:
            table.add_row(*(str(o.get(c, "")) for c in columns))
        sprint(table)



from pathlib import Path

from typer import Option

from patterns.cli.services.deploy import deploy_graph_version
from patterns.cli.services.lookup import IdLookup
from patterns.cli.services.output import sprint, abort_on_error

_graph_help = "The location of the graph.yml file of the graph to deploy"
_graph_version_id_help = "The id of the graph version to deploy"
_environment_help = "The name of the Patterns environment deploy to"
_organization_help = "The name of the Patterns organization that the graph specified with --graph was uploaded to"


def deploy(
    environment: str = Option("", "-e", "--environment", help=_environment_help),
    organization: str = Option("", "-o", "--organization", help=_organization_help),
    graph: Path = Option(None, help=_graph_help),
    graph_version_id: str = Option(""),
):
    """Deploy a previously uploaded graph version

    You can specify either '--graph-version-id' to deploy a specific version, or
    '--graph' to deploy the latest uploaded version of a graph.
    """
    ids = IdLookup(
        environment_name=environment,
        organization_name=organization,
        explicit_graph_path=graph,
        explicit_graph_version_id=graph_version_id,
    )

    with abort_on_error("Deploy failed"):
        deploy_graph_version(ids.graph_version_id, ids.environment_id)

    sprint(f"[success]Graph deployed.")



from requests.models import Response


def download_graph_version(name: str, organization_name: str) -> Response:
    # manifest = GraphManifestBuilder(
    #     directory=pth_to_root, cfg=cfg
    # ).build_manifest_from_config()
    # zipf = compress_directory(pth_to_root)
    # b64_zipf = base64.b64encode(zipf.read())
    # resp = post(
    #     Endpoints.GRAPH_VERSIONS_UPLOAD,
    #     data={"graph_manifest": manifest, "zip": b64_zipf.decode()},
    # )
    # return resp
    pass



from rich.table import Table
from typer import Option

from patterns.cli.config import (
    read_devkit_config,
    write_devkit_config,
    get_devkit_config_path,
)
from patterns.cli.services.environments import get_environment_by_id
from patterns.cli.services.lookup import IdLookup
from patterns.cli.services.organizations import get_organization_by_id
from patterns.cli.services.output import sprint

_config_help = "The name of the organization to use by default"
_environment_help = "The name of the organization to use by default"


def config(
    organization: str = Option("", "-o", "--organization", help=_config_help),
    environment: str = Option("", "-e", "--environment", help=_environment_help),
):
    """Change the default values used by other commands"""
    ids = IdLookup(
        organization_name=organization,
        environment_name=environment,
        ignore_cfg_environment=True,
    )
    if organization:
        ids.cfg.organization_id = ids.organization_id
    if environment:
        ids.cfg.environment_id = ids.environment_id
    write_devkit_config(ids.cfg)

    sprint(f"[info]Your patterns config is located at "
           f"[code]{get_devkit_config_path().as_posix()}")

    t = Table(show_header=False)
    try:
        name = get_organization_by_id(ids.organization_id)["name"]
        t.add_row("organization", name)
    except Exception:
        t.add_row("organization_id", ids.organization_id)
    try:
        name = get_environment_by_id(ids.environment_id)["name"]
        t.add_row("environment", name)
    except Exception:
        t.add_row("environment_id", ids.environment_id)
    if ids.cfg.auth_server:
        t.add_row("auth_server.domain", ids.cfg.auth_server.domain)
        t.add_row("auth_server.audience", ids.cfg.auth_server.audience)
        t.add_row("auth_server.devkit_client_id", ids.cfg.auth_server.devkit_client_id)
    sprint(t)



import os
from typing import Any

import requests
from requests import Response, Session

from patterns.cli.config import (
    read_devkit_config,
    CliConfig,
    write_devkit_config,
    AuthServer,
)
from patterns.cli.services.output import abort, abort_on_error

API_BASE_URL = (
    os.environ.get("PATTERNS_API_URL", "https://api-production.patterns.app/").rstrip("/")
    + "/"
)

AUTH_TOKEN_ENV_VAR = "PATTERNS_AUTH_TOKEN"
AUTH_TOKEN_PREFIX = "JWT"

PUBLIC_API_BASE_URL = "api/v1"


def _get_api_session() -> Session:
    s = requests.Session()
    auth_token = _get_auth_token()
    s.headers.update(
        {
            "Authorization": f"{AUTH_TOKEN_PREFIX} {auth_token}",
            "Accept": "application/json",
        }
    )
    return s


def get_auth_server() -> AuthServer:
    with abort_on_error("Failed getting the auth server"):
        resp = requests.get(API_BASE_URL + Endpoints.TOKEN_AUTHSERVER)
        resp.raise_for_status()
        return AuthServer(**resp.json())


def _get_auth_token() -> str:
    override = os.environ.get(AUTH_TOKEN_ENV_VAR)
    if override:
        return override

    cfg = read_devkit_config()
    if not cfg.token:
        abort("You must be logged in to use this command. Run [code]patterns login")

    with abort_on_error("Failed verifying auth token"):
        resp = requests.post(
            API_BASE_URL + Endpoints.TOKEN_VERIFY, json={"token": cfg.token}
        )
    if resp.status_code == 401:
        if cfg.refresh:
            cfg = _refresh_token(cfg)
    else:
        resp.raise_for_status()
    return cfg.token


def _refresh_token(cfg: CliConfig) -> CliConfig:
    auth_server = cfg.auth_server
    if not auth_server or not cfg.refresh:
        abort("Not logged in\n[info]You can log in with [code]patterns login")

    with abort_on_error("Failed to refresh access token"):
        resp = requests.post(
            f"https://{auth_server.domain}/oauth/token",
            data={
                "client_id": auth_server.devkit_client_id,
                "refresh_token": cfg.refresh,
                "grant_type": "refresh_token",
            },
        )
        resp.raise_for_status()

    cfg.token = resp.json()["access_token"]
    write_devkit_config(cfg)

    return cfg


def get_json(
    path: str,
    params: dict = None,
    session: Session = None,
    base_url: str = API_BASE_URL,
    **kwargs,
) -> Any:
    resp = get(path, params, session, base_url, **kwargs)
    resp.raise_for_status()
    return resp.json()


def get(
    path: str,
    params: dict = None,
    session: Session = None,
    base_url: str = API_BASE_URL,
    **kwargs,
) -> Response:
    session = session or _get_api_session()
    resp = session.get(base_url + path, params=params or {}, **kwargs)
    return resp


def post_for_json(
    path: str,
    json: dict = None,
    session: Session = None,
    base_url: str = API_BASE_URL,
    **kwargs,
) -> Any:
    resp = post(path, json, session, base_url, **kwargs)
    resp.raise_for_status()
    return resp.json()


def post(
    path: str,
    json: dict = None,
    session: Session = None,
    base_url: str = API_BASE_URL,
    **kwargs,
) -> Response:
    session = session or _get_api_session()
    resp = session.post(base_url + path, json=json or {}, **kwargs)
    return resp


class Endpoints:
    TOKEN_CREATE = "auth/jwt/create"
    TOKEN_AUTHSERVER = f"{PUBLIC_API_BASE_URL}/auth/jwt/authserver"
    TOKEN_VERIFY = f"{PUBLIC_API_BASE_URL}/auth/jwt/verify"
    ACCOUNTS_ME = f"{PUBLIC_API_BASE_URL}/accounts/me"
    DEPLOYMENTS_DEPLOY = f"{PUBLIC_API_BASE_URL}/deployments"
    DEPLOYMENTS_TRIGGER_NODE = f"{PUBLIC_API_BASE_URL}/deployments/triggers"
    ENVIRONMENTS_CREATE = f"{PUBLIC_API_BASE_URL}/environments"
    ORGANIZATIONS_LIST = f"{PUBLIC_API_BASE_URL}/organizations"
    EXECUTION_EVENTS = f"{PUBLIC_API_BASE_URL}/nodes/execution_events"
    OUTPUT_DATA = f"{PUBLIC_API_BASE_URL}/nodes/store_data/latest"
    WEBHOOKS = f"{PUBLIC_API_BASE_URL}/webhooks"
    COMPONENTS_LIST = f"{PUBLIC_API_BASE_URL}/marketplace/components"
    COMPONENTS_CREATE = f"{PUBLIC_API_BASE_URL}/marketplace/components/versions"

    @classmethod
    def organization_by_slug(cls, slug: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/organizations/slug/{slug}"

    @classmethod
    def organization_by_id(cls, organization_uid: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/organizations/{organization_uid}"

    @classmethod
    def graphs_list(cls, organization_uid: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/organizations/{organization_uid}/graphs"

    @classmethod
    def graphs_latest(cls, graph_uid: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/graphs/{graph_uid}/latest"

    @classmethod
    def graph_version_download(cls, graph_version_uid: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/graph_versions/{graph_version_uid}/zip"

    @classmethod
    def component_download(cls, organization: str, component: str, version: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/marketplace/components/{organization}/{component}/{version}/zip"

    @classmethod
    def graph_by_slug(cls, organization_uid: str, slug: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/organizations/{organization_uid}/graphs/slug/{slug}"

    @classmethod
    def graph_version_by_id(cls, graph_version_uid: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/graph_versions/{graph_version_uid}"

    @classmethod
    def graph_version_create(cls, organization_uid: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/organizations/{organization_uid}/graph_versions"

    @classmethod
    def environments_list(cls, organization_uid: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/organizations/{organization_uid}/environments"

    @classmethod
    def environment_by_slug(cls, organization_uid: str, slug: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/organizations/{organization_uid}/environments/slug/{slug}"

    @classmethod
    def environment_by_id(cls, environment_uid: str) -> str:
        return f"{PUBLIC_API_BASE_URL}/environments/{environment_uid}"



import json
from pathlib import Path

from patterns.cli.services.api import Endpoints, post_for_json
from patterns.configuration.edit import GraphDirectoryEditor


def upload_graph_version(
    graph_yaml_path: Path, organization_uid: str, add_missing_node_ids: bool
) -> dict:
    editor = GraphDirectoryEditor(graph_yaml_path)
    if add_missing_node_ids:
        editor.add_missing_node_ids()
    payload = {"slug": editor.graph_slug(), "root_yaml_path": editor.yml_path.name}
    return post_for_json(
        Endpoints.graph_version_create(organization_uid),
        data={"payload": json.dumps(payload)},
        files={"file": editor.compress_directory()},
    )



from __future__ import annotations

from enum import Enum

from patterns.cli.services.api import Endpoints, post_for_json


class TypeChoices(str, Enum):
    pubsub = "pubsub"
    http = "http"
    local = "local"


def trigger_node(
    node_id: str,
    graph_version_uid: str,
    environment_uid: str,
    execution_type: TypeChoices = TypeChoices.pubsub,
) -> list[dict]:
    return post_for_json(
        Endpoints.DEPLOYMENTS_TRIGGER_NODE,
        json={
            "node_id": node_id,
            "environment_uid": environment_uid,
            "graph_version_uid": graph_version_uid,
            "execution_type": execution_type.name,
        },
    )



from __future__ import annotations

from patterns.cli.services.api import Endpoints, get_json
from patterns.cli.services.pagination import paginated


@paginated
def paginated_graphs(organization_uid: str):
    return get_json(Endpoints.graphs_list(organization_uid))


@paginated
def paginated_execution_events(environment_uid: str, graph_uid: str, node_id: str):
    return get_json(
        Endpoints.EXECUTION_EVENTS,
        params={
            "environment_uid": environment_uid,
            "graph_uid": graph_uid,
            "node_id": node_id,
        },
    )


@paginated
def paginated_output_data(
    environment_uid: str, graph_uid: str, store_name: str
):
    return get_json(
        Endpoints.OUTPUT_DATA,
        params={
            "environment_uid": environment_uid,
            "graph_uid": graph_uid,
            "store_name": store_name,
        },
    )


@paginated
def paginated_webhook_urls(
    environment_uid: str, graph_uid: str,
):
    return get_json(
        Endpoints.WEBHOOKS,
        params={"environment_uid": environment_uid, "graph_uid": graph_uid},
    )


@paginated
def paginated_graph_components():
    """All public components"""
    return get_json(Endpoints.COMPONENTS_LIST)



from __future__ import annotations

import base64
import dataclasses
import hashlib
import os
import re
import urllib.parse
from socketserver import BaseRequestHandler
from urllib.parse import ParseResult

import requests
from requests import HTTPError

from patterns.cli.config import update_devkit_config
from patterns.cli.services.api import (
    AuthServer,
    get_auth_server,
)
from patterns.cli.services.auth import (
    LOCAL_OAUTH_PORT,
    BaseOAuthRequestHandler,
    execute_oauth_flow,
)


def login():
    auth_server = get_auth_server()

    # The code_verifier and code_challenge are part of the OAuth PKCE spec.
    # https://datatracker.ietf.org/doc/html/rfc7636#section-4.1
    # https://developer.okta.com/blog/2019/08/22/okta-authjs-pkce
    #
    # We make some random bits (code_verifier), hash it (code_challenge) and send
    # the challenge with our initial authorize request.  When the user is redirected
    # back, we post the unhashed value (code_verifier) when obtaining the
    # token.  The server then hashes the code_verifier to match the initial value that
    # was sent by the redirected client.
    #
    # This flow stands in the place of having a client secret, since the devkit is
    # Open Source we cannot use client secret based auth.
    #
    # The encoded verifier only needs 32 bytes of entropy, but we use 33 because
    # otherwise we end up with padding on our bas64 encoded value
    code_verifier = base64.urlsafe_b64encode(os.urandom(33)).decode("utf-8")

    code_challenge = hashlib.sha256(code_verifier.encode("utf-8")).digest()
    code_challenge = base64.urlsafe_b64encode(code_challenge).decode("utf-8")
    # Remove padding (per Auth0 samples)
    code_challenge = code_challenge.replace("=", "")

    # Random state is sent with the initial request and received on the redirect
    # and is supposed to mitigate CSRF attacks:
    # https://auth0.com/docs/secure/attack-protection/state-parameters
    state = os.urandom(50).hex()

    redirect_url = (
        f"http://localhost:{LOCAL_OAUTH_PORT}{LoginRequestHandler.handled_path}"
    )

    params = {
        "response_type": "code",
        "code_challenge_method": "S256",
        "code_challenge": code_challenge,
        "client_id": auth_server.devkit_client_id,
        "redirect_uri": redirect_url,
        "scope": "profile email openid offline_access",
        "audience": auth_server.audience,
        "state": state,
    }
    query = urllib.parse.urlencode(params)
    url = f"https://{auth_server.domain}/authorize?{query}"

    login_config = LoginConfig(
        auth_server=auth_server,
        state=state,
        code_verifier=code_verifier,
        redirect_url=redirect_url,
    )

    def on_request(handler: BaseRequestHandler):
        handler._login_config = login_config

    execute_oauth_flow(url, LoginRequestHandler, on_request)


@dataclasses.dataclass
class LoginConfig:
    auth_server: AuthServer
    state: str
    code_verifier: str
    redirect_url: str


class LoginRequestHandler(BaseOAuthRequestHandler):
    handled_path: str = "/auth_callback"
    _login_config: LoginConfig = None

    def handle_callback(self, parsed_url: ParseResult):
        qs = urllib.parse.parse_qs(parsed_url.query)
        if not (code := self.get_single_queryparam("code", qs)):
            return

        if not (state := self.get_single_queryparam("state", qs)):
            return

        login_config = self._login_config
        expected_state = login_config.state
        if state != expected_state:
            return self.finish_with_error(
                401,
                f"An invalid state was returned.  Expected {expected_state} but was {state}.  Unable to login",
            )

        # Exchange code for access & refresh tokens
        response = requests.post(
            f"https://{login_config.auth_server.domain}/oauth/token",
            json={
                "client_id": login_config.auth_server.devkit_client_id,
                "code_verifier": login_config.code_verifier,
                "code": code,
                "grant_type": "authorization_code",
                "redirect_uri": login_config.redirect_url,
            },
        )
        response.raise_for_status()

        json = response.json()
        if "refresh_token" not in json or "access_token" not in json:
            return self.finish_with_error(
                401, f"We did not receive a valid authorization result: {json}"
            )

        update_devkit_config(
            auth_server=login_config.auth_server,
            refresh=json["refresh_token"],
            token=json["access_token"],
        )

        return self.finish_with_success(
            "Login successful", "Successfully logged in!  You may close this window."
        )



from __future__ import annotations

from datetime import date, datetime
from typing import Callable

import requests

try:
    from ratelimit import limits, sleep_and_retry
except ImportError:
    limits = None
    sleep_and_retry = None
from requests import Response


class JsonHttpApiConnection:
    def __init__(
        self,
        default_params: dict = None,
        default_headers: dict = None,
        date_format: str = "%F %T",
        raise_for_status: bool = True,
        ratelimit_calls_per_min: int = 1000,
        remove_none_params: bool = True,
        method: str = "get",
    ):
        self.default_params = default_params or {}
        self.default_headers = default_headers or {}
        self.date_format = date_format
        self.raise_for_status = raise_for_status
        self.ratelimit_calls_per_min = ratelimit_calls_per_min
        self.g = self.add_rate_limiting(self.get)
        self.remove_none_params = remove_none_params
        self.method = method

    def add_rate_limiting(self, f: Callable):
        if limits is None:
            return f
        g = sleep_and_retry(f)
        g = limits(calls=self.ratelimit_calls_per_min, period=60)(g)
        return g

    def get_default_params(self) -> dict:
        return self.default_params.copy()

    def get_default_headers(self) -> dict:
        return self.default_headers.copy()

    def validate_params(self, params: dict) -> dict:
        formatted = {}
        for k, v in params.items():
            if self.remove_none_params and v is None:
                continue
            if isinstance(v, datetime) or isinstance(v, date):
                v = v.strftime(self.date_format)
            formatted[k] = v
        return formatted

    def get(
        self, url: str, params: dict = None, headers: dict = None, method=None, **kwargs
    ) -> Response:
        method = method or self.method
        default_params = self.get_default_params()
        if params:
            default_params.update(params)
        final_headers = self.get_default_headers()
        if headers:
            final_headers.update(headers)
        final_params = self.validate_params(default_params)
        if method == "get":
            resp = requests.get(
                url, params=final_params, headers=final_headers, **kwargs
            )
        elif method == "post":
            final_headers["Content-Type"] = final_headers.get(
                "Content-Type", "application/json"
            )
            resp = requests.post(
                url, json=final_params, headers=final_headers, **kwargs
            )
        else:
            raise NotImplementedError(self.method)
        if self.raise_for_status:
            resp.raise_for_status()
        return resp

    def post(self, *args, **kwargs):
        return self.get(*args, **kwargs, method="post")



from __future__ import annotations

from patterns.cli.services.api import Endpoints, get_json
from patterns.cli.services.pagination import paginated


def get_environment_by_name(organization_uid: str, name: str) -> dict:
    return get_json(Endpoints.environment_by_slug(organization_uid, name))


def get_environment_by_id(environment_uid: str) -> dict:
    return get_json(Endpoints.environment_by_id(environment_uid))


@paginated
def paginated_environments(environment_uid: str):
    return get_json(Endpoints.environments_list(environment_uid))



from __future__ import annotations

import functools
import io
import re
from io import StringIO
from pathlib import Path
from typing import List, Dict, Any, Optional, Union, IO, Iterator, Callable
from zipfile import ZipFile, ZipInfo

import ruyaml

from patterns.cli.helpers import compress_directory, random_node_id
from patterns.cli.services.graph import resolve_graph_path

MISSING = object()


class GraphConfigEditor:
    """Edit a graph.yml file, preserving comments

    By default, constructing an instance of this class will raise an exception if the
    yaml file doesn't exist or can't be parsed.

    If you pass `read=False` to the constructor, the file won't be read, and you'll
    start with an empty config.

    You can also pass `path_to_graph_yml=None` if you aren't going to write back to
    disk.
    """

    def __init__(self, path_to_graph_yml: Optional[Path], read: bool = True):
        self._yaml = ruyaml.YAML()
        self._path_to_graph_yml = path_to_graph_yml
        self._yaml.indent(mapping=2, sequence=4, offset=2)
        # read text manually instead of loading the Path directly to normalize line
        # breaks. Ruyaml opens files in binary mode (bypassing universal newline
        # support), then proceeds to behave incorrectly in the presence of \r\n, adding
        # extra line breaks in the output.
        if read:
            with self._path_to_graph_yml.open() as f:
                text = f.read()
            self._cfg = self._yaml.load(text) or {}
            # ruyaml doesn't provide a way to preserve indentation,
            # so pick a value that matches the least-indented list item we see.
            matches = (
                len(m.group(1)) for m in re.finditer(r"^( *)-", text, re.MULTILINE)
            )
            # ruyaml's indent number includes the length  of "- " for some reason
            indent = min(matches, default=2) + 2
        else:
            self._cfg = {}
            indent = 4

        self._yaml.indent(
            mapping=int(indent / 2), sequence=indent, offset=max(0, indent - 2)
        )

    def write(self):
        """Write the config back to the file"""
        self._yaml.dump(self._cfg, self._path_to_graph_yml)

    def dump(self) -> str:
        """Return the edited config as a yaml string"""
        s = StringIO()
        self._yaml.dump(self._cfg, s)
        return s.getvalue()

    def set_name(self, name: str) -> GraphConfigEditor:
        self._cfg["title"] = name
        return self

    def get_title(self) -> Optional[str]:
        return self._cfg.get("title")

    def get_slug(self) -> Optional[str]:
        """Return the slug value listed in the yaml, or None if it's not present"""
        return self._cfg.get("slug")

    def add_function_node_dict(self, node: dict) -> GraphConfigEditor:
        d = {k: v for (k, v) in node.items() if v is not None}
        for k in ("node_file", "id", "webhook"):
            if (
                k in d
                and d[k]
                and any(it.get(k) == d[k] for it in self._cfg.get("functions", []))
            ):
                raise ValueError(
                    f"{k} '{d[k]}' already defined in the graph configuration"
                )

        if "functions" not in self._cfg:
            self._cfg["functions"] = []
        self._cfg["functions"].append(d)
        return self

    def add_node(
        self,
        node_file: str,
        schedule: str = None,
        inputs: Dict[str, str] = None,
        outputs: Dict[str, str] = None,
        parameters: Dict[str, Any] = None,
        title: str = None,
        id: Optional[str] = MISSING,
        description: str = None,
        description_file: str = None,
    ) -> GraphConfigEditor:
        if id is MISSING:
            id = random_node_id()
        self.add_function_node_dict(
            {
                "node_file": node_file,
                "schedule": schedule,
                "inputs": inputs,
                "outputs": outputs,
                "parameters": parameters,
                "title": title,
                "id": str(id) if id else id,
                "description": description,
                "description_file": description_file,
            }
        )
        return self

    def add_store(
        self,
        name: str,
        table: bool,
        title: str = None,
        id: Optional[str] = MISSING,
        schema: str = None,
    ):
        if id is MISSING:
            id = random_node_id()
        d = {
            "table" if table else "stream": name,
            "title": title,
            "id": str(id) if id else id,
            "schema": schema,
        }
        d = {k: v for (k, v) in d.items() if v is not None}

        for k in ("table", "stream", "id"):
            if d.get(k) and any(
                it.get(k) == d[k] for it in self._cfg.get("stores", [])
            ):
                raise ValueError(
                    f"{k} '{d[k]}' already defined in the graph configuration"
                )

        if "stores" not in self._cfg:
            self._cfg["stores"] = []
        self._cfg["stores"].append(d)
        return self

    def add_webhook(
        self,
        webhook: str,
        title: str = None,
        id: Optional[str] = MISSING,
        description: str = None,
        description_file: str = None,
    ) -> GraphConfigEditor:
        if id is MISSING:
            id = random_node_id()
        self.add_function_node_dict(
            {
                "webhook": webhook,
                "title": title,
                "id": str(id) if id else id,
                "description": description,
                "description_file": description_file,
            }
        )
        return self

    def add_component_uses(
        self,
        component_key: str,
        schedule: str = None,
        inputs: Dict[str, str] = None,
        outputs: Dict[str, str] = None,
        parameters: Dict[str, Any] = None,
        title: str = None,
        id: Optional[str] = MISSING,
        description: str = None,
        description_file: str = None,
    ) -> GraphConfigEditor:
        if id is MISSING:
            id = random_node_id()
        self.add_function_node_dict(
            {
                "uses": component_key,
                "schedule": schedule,
                "inputs": inputs,
                "outputs": outputs,
                "parameters": parameters,
                "title": title,
                "id": str(id) if id else id,
                "description": description,
                "description_file": description_file,
            }
        )
        return self

    def add_missing_node_ids(self) -> GraphConfigEditor:
        """Add a random id to any node entry that doesn't specify one"""
        for node in self.all_nodes():
            if "id" not in node:
                node["id"] = random_node_id()
        return self

    def all_nodes(self) -> Iterator[dict]:
        """Return an iterator over all function and store node declarations"""
        yield from self.function_nodes()
        yield from self.store_nodes()

    def function_nodes(self) -> Iterator[dict]:
        nodes = self._cfg.get("functions")
        if not isinstance(nodes, list):
            return

        for node in nodes:
            if not isinstance(node, dict):
                continue
            yield node

    def store_nodes(self) -> Iterator[dict]:
        nodes = self._cfg.get("stores")
        if not isinstance(nodes, list):
            return

        for node in nodes:
            if not isinstance(node, dict):
                continue
            yield node


class GraphDirectoryEditor:
    def __init__(self, graph_path: Path, overwrite: bool = False):
        """
        :param graph_path: The path to a graph.yml file, or a directory containing one
        :param overwrite: If False, raise an exception in add_node_from_zip if a node
            exists and differs from the extracted content.
        """
        try:
            self.yml_path = resolve_graph_path(graph_path, exists=True)
        except ValueError:
            self.yml_path = resolve_graph_path(graph_path, exists=False)
        self.dir = self.yml_path.parent
        self.overwrite = overwrite
        if self.yml_path.is_file():
            self._cfg = self._editor(self.yml_path)
        else:
            self._cfg: Optional[GraphConfigEditor] = None

    def graph_slug(self) -> str:
        """Return the graph name slug based on the graph directory name"""
        if self._cfg and self._cfg.get_slug():
            name = self._cfg.get_slug()
        else:
            name = self.yml_path.parent.name
        return re.sub(r"[^a-zA-Z0-9]", "-", name)

    def compress_directory(self) -> io.BytesIO:
        """Return an in-memory zip file containing the compressed graph directory"""
        return compress_directory(self.dir)

    def add_node_from_file(self, dst_path: Union[Path, str], file: IO[bytes]):
        """Write the content of a file to dst_path

       :param dst_path: Path relative to the output graph directory
       :param file: A file-like object open in read mode
       """
        dst_path = Path(dst_path)
        self._write_file(dst_path, file)
        self._add_cfg_node(dst_path)

    def add_node_from_zip(
        self,
        src_path: Union[Path, str],
        dst_path: Union[Path, str],
        zf: Union[ZipFile, Path, IO[bytes]],
    ) -> GraphDirectoryEditor:
        """Copy the node or subgraph located at src_path in zipfile to dst_path

        :param src_path: Path relative to the root of zipfile
        :param dst_path: Path relative to the output graph directory
        :param zf: A ZipFile open in read mode, or a path to a zip file to open
        """
        src_path = Path(src_path)
        dst_path = Path(dst_path)
        if isinstance(zf, ZipFile):
            self._add(src_path, dst_path, zf)
        else:
            with ZipFile(zf, "r") as f:
                self._add(src_path, dst_path, f)
        return self

    def add_missing_node_ids(self) -> GraphDirectoryEditor:
        """Add a random id to any node entry that doesn't specify one

        This will update all graph.yml files in the directory
        """
        for editor in self._graph_editors():
            editor.add_missing_node_ids()
            editor.write()
        return self

    def _add(self, src_path: Path, dst_path: Path, zf: ZipFile):
        if src_path.name == "graph.yml":

            def dirname(p):
                if len(p.parts) <= 1:
                    return ""
                return _zip_name(p.parent) + "/"

            src_dir = dirname(src_path)
            dst_dir = dirname(dst_path)

            for info in zf.infolist():
                if info.filename.startswith(src_dir) and not info.is_dir():
                    new_name = dst_dir + info.filename[len(src_dir) :]
                    self._extract_file(info, Path(new_name), zf)
        else:
            self._extract_file(zf.getinfo(_zip_name(src_path)), dst_path, zf)
        self._add_cfg_node(dst_path)

    def _add_cfg_node(self, dst_path: Path):
        if not self._cfg or str(dst_path) == "graph.yml":
            return
        try:
            self._cfg.add_node(_zip_name(dst_path)).write()
        except ValueError:
            pass  # node already exists, leave it unchanged

    def _extract_file(self, member: ZipInfo, dst_path: Path, zf: ZipFile):
        full_dst_path = self.dir / dst_path
        if full_dst_path.is_dir():
            raise ValueError(
                f"Cannot extract {dst_path}: a directory by that name exists"
            )
        if self.overwrite or not full_dst_path.is_file():
            member.filename = _zip_name(dst_path)
            zf.extract(member, self.dir)
        else:
            with zf.open(member, "r") as f:
                self._write_file(dst_path, f)

    def _write_file(self, dst_path: Path, file: IO[bytes]):
        full_dst_path = self.dir / dst_path
        new_content = io.TextIOWrapper(file).read()

        if not self.overwrite:
            try:
                old_content = full_dst_path.read_text()
            except FileNotFoundError:
                pass
            else:
                if new_content != old_content:
                    raise FileOverwriteError(
                        full_dst_path,
                        f"Cannot extract {dst_path}: would overwrite existing file",
                    )
        full_dst_path.write_text(new_content)

    def _graph_editors(self) -> Iterator[GraphConfigEditor]:
        for p in self.dir.rglob("graph.yml"):
            yield self._editor(p)

    @functools.lru_cache(maxsize=None)
    def _editor(self, yaml_path: Path) -> GraphConfigEditor:
        return GraphConfigEditor(yaml_path)


class FileOverwriteError(Exception):
    def __init__(self, file_path: Path, message: str) -> None:
        super().__init__(message)
        self.file_path = file_path


def _zip_name(p: Path):
    return "/".join(p.parts)



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator, Any

try:
    # DataFrame is only used in type annotations
    from pandas import DataFrame
except ImportError:
    DataFrame = None


class TableVersion:
    @property
    def name(self) -> str:
        ...

    @property
    def storage(self):
        ...

    @property
    def schema(self):
        ...

    @property
    def record_count(self) -> int | None:
        ...


class InputTableMethods:
    @classmethod
    def read(
        cls,
        as_format: str = "records",
        chunksize: int | None = None,
    ) -> list[dict] | DataFrame | Iterator[list[dict]] | Iterator[DataFrame]:
        ...

    @classmethod
    def read_sql(
        cls,
        sql: str,
        as_format: str = "records",
        chunksize: int | None = None,
    ) -> list[dict] | DataFrame | Iterator[list[dict]] | Iterator[DataFrame]:
        ...

    @classmethod
    def get_active_version(cls) -> TableVersion | None:
        ...

    @classmethod
    def has_active_version(cls) -> bool:
        ...


class OutputTableMethods:
    @classmethod
    def write(cls, records: DataFrame | list[dict] | dict, replace=False):
        ...

    @classmethod
    def execute_sql(cls, sql: str, created: bool = False):
        ...

    @classmethod
    def create_new_version(cls) -> TableVersion:
        ...

    @classmethod
    def get_active_version(cls) -> TableVersion | None:
        ...

    @classmethod
    def set_active_version(cls, table: TableVersion):
        ...

    @classmethod
    def has_active_version(cls) -> bool:
        ...

    @classmethod
    def signal_create(cls):
        ...

    @classmethod
    def signal_update(cls):
        ...

    @classmethod
    def signal_reset(cls):
        ...


class InputStreamMethods:
    @classmethod
    def consume_records(cls, with_metadata: bool = False) -> Iterator[dict]:
        ...

    def __iter__(cls) -> Iterator[dict]:
        ...

    @classmethod
    def checkpoint(cls):
        ...

    @classmethod
    def rollback(cls, records_to_rollback: int = 1):
        ...

    @classmethod
    def consume_all(cls) -> Iterator[str | None]:
        ...


class OutputStreamMethods:
    @classmethod
    def write(cls, record: DataFrame | list[dict] | dict):
        ...


class StateMethods:
    @classmethod
    def set(cls, state: dict):
        ...

    @classmethod
    def set_value(cls, key: str, value: Any):
        ...

    @classmethod
    def get(cls) -> dict:
        ...

    @classmethod
    def get_value(cls, key: str) -> Any:
        ...

    @classmethod
    def should_continue(
        cls, pct_of_limit: float = None, seconds_till_limit: int = None
    ) -> bool:
        ...

    @classmethod
    def request_new_run(
        cls, trigger_downstream: bool = True, wait_atleast_seconds: int = None
    ):
        ...


class ParameterMethods:
    pass



from .node.node import (
    node,
    InputTable,
    OutputTable,
    InputStream,
    OutputStream,
    Parameter,
    State,
    Stream,
    Table,
)


__all__ = [
    "node",
    "InputTable",
    "OutputTable",
    "InputStream",
    "OutputStream",
    "Parameter",
    "State",
    "Table",
    "Stream",
]



from pathlib import Path


def resolve_graph_path(
    path: Path, exists: bool, create_parents_if_necessary: bool = True
) -> Path:
    """Resolve an explicitly given graph location to a yaml"""
    if path.is_dir():
        f = path / "graph.yml"
        if f.is_file():
            if exists:
                return f.absolute()
            raise ValueError(f"File '{f}' already exists")
        if exists:
            raise ValueError(f"File '{f}' does not exist")
        return f.absolute()
    if path.suffix and path.suffix not in (".yml", ".yaml"):
        raise ValueError(f"Invalid graph file name: {path.name}")
    if path.is_file():
        if not exists:
            raise ValueError(f"Graph '{path}' already exists")
        return path.absolute()
    if exists:
        raise ValueError(f"Graph '{path}' does not exist")
    if path.suffix:
        if create_parents_if_necessary:
            path.parent.mkdir(parents=True)
        return path.absolute()
    if create_parents_if_necessary:
        path.mkdir(parents=True)
    graph_path = (path / "graph.yml").absolute()
    return graph_path




from __future__ import annotations

import urllib.parse
from urllib.parse import ParseResult

from patterns.cli.config import read_devkit_config, update_devkit_config
from patterns.cli.services.auth import (
    LOCAL_OAUTH_PORT,
    BaseOAuthRequestHandler,
    execute_oauth_flow,
)


def logout():
    cfg = read_devkit_config()
    if not cfg.auth_server:
        return

    params = {
        "client_id": cfg.auth_server.devkit_client_id,
        "returnTo": f"http://localhost:{LOCAL_OAUTH_PORT}{LogoutRequestHandler.handled_path}",
    }

    query = urllib.parse.urlencode(params)
    url = f"https://{cfg.auth_server.domain}/v2/logout?{query}"

    execute_oauth_flow(url, LogoutRequestHandler)


class LogoutRequestHandler(BaseOAuthRequestHandler):
    handled_path: str = "/logout_callback"

    def handle_callback(self, parsed_url: ParseResult):
        update_devkit_config(refresh=None, token=None, auth_server=None)
        self.finish_with_success(
            "Successfully logged out", "You have successfully logged out"
        )



import io
import re
from pathlib import Path
from zipfile import ZipFile

import typer
from typer import Option, Argument

from patterns.cli.helpers import random_node_id
from patterns.cli.services.graph import resolve_graph_path
from patterns.cli.services.list import paginated_graph_components
from patterns.cli.services.lookup import IdLookup
from patterns.cli.services.output import abort, prompt_path, abort_on_error
from patterns.cli.services.output import sprint
from patterns.cli.services.pull import download_graph_zip
from patterns.configuration.edit import GraphConfigEditor
from patterns.configuration.edit import GraphDirectoryEditor

create = typer.Typer(name="create", help="Create a graph new or node")

_name_help = "The name of the graph. The location will be used as a name by default"


@create.command()
def graph(
    name: str = Option("", "--name", "-n", help=_name_help),
    location: Path = Argument(None, metavar="GRAPH"),
):
    """Add a new node to a graph"""
    if not location:
        prompt = (
            "Enter a name for the new graph directory [prompt.default](e.g. my_graph)"
        )
        location = prompt_path(prompt, exists=False)
    with abort_on_error("Error creating graph"):
        path = resolve_graph_path(location, exists=False)
    name = name or location.stem
    GraphConfigEditor(path, read=False).set_name(name).write()

    sprint(f"\n[success]Created graph [b]{name}")
    sprint(
        f"\n[info]You can add nodes with [code]cd {location}[/code],"
        f" then [code]patterns create node[/code]"
    )


_graph_help = "The graph to add this node to"
_title_help = "The title of the node. The location will be used as a title by default"
_component_help = "The name of component to use to create this node"


@create.command()
def node(
    title: str = Option("", "--title", "-n", help=_name_help),
    component: str = Option("", "-c", "--component", help=_component_help),
    location: Path = Argument(None),
):
    """Add a new node to a graph

    patterns create node --name='My Node' mynode.py
    """
    if component and location:
        abort("Specify either a component or a node location, not both")

    if component:
        ids = IdLookup(find_nearest_graph=True)
        GraphConfigEditor(ids.graph_file_path).add_component_uses(
            component_key=component
        ).write()
        sprint(f"[success]Added component {component} to graph")
        return

    if not location:
        sprint("[info]Nodes can be python files like [code]ingest.py")
        sprint("[info]Nodes can be sql files like [code]aggregate.sql")
        sprint("[info]You also can add a subgraph like [code]processor/graph.yml")
        message = "Enter a name for the new node file"
        location = prompt_path(message, exists=False)

    if location.exists():
        abort(f"Cannot create node: {location} already exists")

    ids = IdLookup(node_file_path=location, find_nearest_graph=True)
    # Update the graph yaml
    node_file = "/".join(location.absolute().relative_to(ids.graph_directory).parts)
    node_title = title or (
        location.parent.name if location.name == "graph.yml" else location.stem
    )
    with abort_on_error("Adding node failed"):
        editor = GraphConfigEditor(ids.graph_file_path)
        editor.add_node(
            title=node_title, node_file=node_file, id=str(random_node_id()),
        )

    # Write to disk last to avoid partial updates
    if location.suffix == ".py":
        fun_name = re.sub(r"[^a-zA-Z0-9_]", "_", location.stem)
        if re.match(r"\d", fun_name):
            fun_name = f"node_{fun_name}"
        location.write_text(_PY_FILE_TEMPLATE.format(fun_name))
    elif location.suffix == ".sql":
        location.write_text(_SQL_FILE_TEMPLATE)
    elif location.name == "graph.yml":
        location.parent.mkdir(exist_ok=True, parents=True)
        GraphConfigEditor(location, read=False).set_name(node_title).write()
    else:
        abort("Node file must be graph.yml or end in .py or .sql")
    editor.write()

    sprint(f"\n[success]Created node [b]{location}")
    sprint(
        f"\n[info]Once you've edited the node and are ready to run the graph, "
        f"use [code]patterns upload"
    )


_webhook_name_help = "The name of the webhook output stream"


@create.command()
def webhook(
    explicit_graph: Path = Option(None, "--graph", "-g", exists=True, help=_graph_help),
    name: str = Argument(..., help=_webhook_name_help),
):
    """Add a new webhook node to a graph"""
    ids = IdLookup(explicit_graph_path=explicit_graph)

    with abort_on_error("Adding webhook failed"):
        editor = GraphConfigEditor(ids.graph_file_path)
        editor.add_webhook(name, id=random_node_id())
        editor.write()

    sprint(f"\n[success]Created webhook [b]{name}")
    sprint(
        f"\n[info]Once you've deployed the graph, use "
        f"[code]patterns list webhooks[/code] to get the url of the webhook"
    )


_PY_FILE_TEMPLATE = """
from patterns import *


@node
def {}(
    # Declare the node inputs and outputs here:
    # input_stream=InputStream,
    # output_table=OutputTable,
    # myparam=Parameter(type='text'),
):
    # use the inputs and outputs here:
    pass
"""

_SQL_FILE_TEMPLATE = """
create table {{ OutputTable("my_output_table") }}
select
    *
from {{ InputTable("other_node") }}
limit {{ Parameter("limit", "int", default=100) }}
"""



from dataclasses import dataclass
from dataclasses import dataclass
from typing import Callable, Any, TypeVar, Type

from patterns.node._methods import (
    InputTableMethods,
    OutputTableMethods,
    InputStreamMethods,
    OutputStreamMethods,
    StateMethods,
)


def _mixin_attrs():
    def init(self, param_name):
        self.param_name = param_name

    return {"__init__": init}


# In order to support type inference with both `Parameter` and `Parameter()` forms, we
# need to use the metaclass mechanisms to return a type rather than a value when the
# classes are invoked. We add __init__ to the generated classes so that invoking them
# generates new class objects with the description etc. set. This does mean that
# invoking the class will recursively produce more classes and never an actual instance,
# but that fine for how we're using them.
class _InputMeta(type):
    def __new__(
        mcs,
        description: str = None,
        schema: str = None,
        required: bool = True,
    ):
        return super().__new__(mcs, mcs.__name__, (mcs,), _mixin_attrs())

    # noinspection PyMissingConstructor
    def __init__(
        cls,
        description: str = None,
        schema: str = None,
        required: bool = True,
    ):
        cls.description = description
        cls.schema = schema
        cls.required = required


class _OutputMeta(type):
    def __new__(
        mcs,
        description: str = None,
        schema: str = None,
    ):
        return super().__new__(mcs, mcs.__name__, (mcs,), _mixin_attrs())

    # noinspection PyMissingConstructor
    def __init__(
        cls,
        description: str = None,
        schema: str = None,
    ):
        cls.description = description
        cls.schema = schema


class _StateMeta(type):
    def __new__(mcs):
        return super().__new__(mcs, mcs.__name__, (mcs,), _mixin_attrs())

    # noinspection PyMissingConstructor
    def __init__(cls):
        pass


class InputTable(_InputMeta, InputTableMethods):
    pass


class OutputTable(_OutputMeta, OutputTableMethods):
    pass


class InputStream(_InputMeta, InputStreamMethods):
    pass


class OutputStream(_OutputMeta, OutputStreamMethods):
    pass


class Table(InputTableMethods, OutputTableMethods):
    def __init__(
        self,
        name: str,
        mode: str = "r",
        description: str = None,
        schema: str = None,
        required: bool = True,
    ):
        pass


class Stream(InputStreamMethods, OutputStreamMethods):
    def __init__(
        self,
        name: str,
        mode: str = "r",
        description: str = None,
        schema: str = None,
        required: bool = True,
    ):
        pass


class State(_StateMeta, StateMethods):
    pass


T = TypeVar("T")


class _Parameter(str):
    def __call__(
        self,
        description: str = None,
        type: Type[T] = str,
        default: Any = None,
    ) -> T:
        pass


Parameter = _Parameter()


@dataclass(frozen=True)
class NodeFunction:
    function: Callable


def node(function: Callable):
    """A decorator that registers a function to execute when a node runs"""
    return NodeFunction(function)



from patterns.cli.config import (
    update_devkit_config,
    get_devkit_config_path,
)
from patterns.cli.services import login as login_service
from patterns.cli.services.accounts import me
from patterns.cli.services.lookup import IdLookup
from patterns.cli.services.output import sprint, abort_on_error


def login():
    """Log in to your Patterns account"""

    with abort_on_error("Login failed"):
        login_service.login()

    ids = IdLookup(ignore_local_cfg=True)
    with abort_on_error("Fetching account failed"):
        update_devkit_config(
            organization_id=ids.organization_id, environment_id=ids.environment_id
        )

    with abort_on_error("Fetching user profile failed"):
        profile = me()

    sprint(
        f"\n[success]Logged in to Patterns organization [b]{ids.organization_name}[/b] "
        f"as [b]{profile['username']}[/b] ([b]{profile['email']}[/b])"
    )
    sprint(
        f"\n[info]Your login information is stored at "
        f"{get_devkit_config_path().as_posix()}"
    )
    sprint(
        f"\n[info]If you want to create a new graph, run "
        f"[code]patterns create graph[/code] get started"
    )



from __future__ import annotations

import re

from patterns.cli.services.api import Endpoints, get

COMPONENT_RE = re.compile(r"([\w\-]+)/([\w\d\-]+)@([\w\d\-.]+)")


def download_graph_zip(graph_version_uid: str) -> bytes:
    resp = get(Endpoints.graph_version_download(graph_version_uid))
    resp.raise_for_status()
    return resp.content


def download_component_zip(component_key: str) -> bytes:
    org, comp, v = COMPONENT_RE.fullmatch(component_key).groups()
    resp = get(Endpoints.component_download(org, comp, v))
    resp.raise_for_status()
    return resp.content



from pathlib import Path

from typer import Option, Argument

from patterns.cli.services.deploy import deploy_graph_version
from patterns.cli.services.graph_components import create_graph_component
from patterns.cli.services.lookup import IdLookup
from patterns.cli.services.output import sprint, abort_on_error
from patterns.cli.services.upload import upload_graph_version

_graph_help = "The location of the graph.yml file for the graph to upload"
_deploy_help = "Whether or not to automatically deploy the graph after upload"
_organization_help = "The name of the Patterns organization to upload to"
_environment_help = "The name of the Patterns environment to use if deploying the graph"
_component_help = "After uploading, publish the graph version as a public component"


def upload(
    deploy: bool = Option(True, "--deploy/--no-deploy", help=_deploy_help),
    organization: str = Option("", "-o", "--organization", help=_organization_help),
    environment: str = Option("", "-e", "--environment", help=_environment_help),
    graph: Path = Argument(None, exists=True, help=_graph_help),
    publish_component: bool = Option(False, help=_component_help),
):
    """Upload a new version of a graph to Patterns"""
    ids = IdLookup(
        environment_name=environment,
        organization_name=organization,
        explicit_graph_path=graph,
    )

    with abort_on_error("Upload failed"):
        resp = upload_graph_version(
            ids.graph_file_path,
            ids.organization_id,
            add_missing_node_ids=not publish_component,
        )

    graph_version_id = resp["uid"]
    ui_url = resp["ui_url"]
    sprint(f"\n[success]Uploaded new graph version with id [b]{graph_version_id}")
    if resp.get("errors"):
        sprint(f"[error]Graph contains the following errors:")
        for error in resp["errors"]:
            sprint(f"\t[error]{error}")

    if publish_component:
        with abort_on_error("Error creating component"):
            resp = create_graph_component(graph_version_id)
            resp_org = resp["organization"]["slug"]
            resp_versions = resp["version_names"]
            resp_component = resp["component"]["slug"]
            resp_id = resp["uid"]
            sprint(
                f"[success]Published graph component "
                f"[b]{resp_org}/{resp_component}[/b] "
                f"with versions [b]{resp_versions}[/b] "
                f"at id [b]{resp_id}"
            )
    elif deploy:
        with abort_on_error("Deploy failed"):
            deploy_graph_version(graph_version_id, ids.environment_id)
        sprint(f"[success]Graph deployed")

    sprint(f"\n[info]Visit [code]{ui_url}[/code] to view your graph")



import os
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Optional

import yaml

from patterns.cli.config import (
    read_devkit_config,
    CliConfig,
    update_devkit_config,
)
from patterns.cli.services.environments import (
    get_environment_by_name,
    paginated_environments,
)
from patterns.cli.services.graph import resolve_graph_path
from patterns.cli.services.graph_versions import (
    get_graph_by_slug,
    get_latest_graph_version,
    get_graph_version_by_id,
)
from patterns.cli.services.organizations import (
    get_organization_by_name,
    paginated_organizations,
)
from patterns.cli.services.output import prompt_choices, prompt_path

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


@dataclass
class IdLookup:
    """A cached lookup of various ids from whatever info you can give it."""

    environment_name: str = None
    organization_name: str = None
    explicit_graph_path: Path = None
    node_file_path: Path = None
    explicit_graph_version_id: str = None
    ignore_local_cfg: bool = False
    ignore_cfg_environment: bool = False
    explicit_graph_name: str = None
    find_nearest_graph: bool = False

    @cached_property
    def organization_id(self) -> str:
        if self.organization_name:
            return get_organization_by_name(self.organization_name)["uid"]
        if self.cfg.organization_id:
            return self.cfg.organization_id
        organizations = list(paginated_organizations())
        orgs_by_name = {org["name"]: org for org in organizations}
        if len(organizations) == 1:
            org = organizations[0]
            update_devkit_config(organization_id=org["uid"])
        else:
            org_name = prompt_choices(
                "Available organizations",
                "Select an organization",
                choices=orgs_by_name.keys(),
            )
            org = orgs_by_name[org_name]
        self.organization_name = org["name"]
        return org["uid"]

    @cached_property
    def environment_id(self) -> str:
        if self.environment_name:
            env = get_environment_by_name(self.organization_id, self.environment_name)
            return env["uid"]
        if self.cfg.environment_id and not self.ignore_cfg_environment:
            return self.cfg.environment_id

        environments = list(paginated_environments(self.organization_id))
        envs_by_name = {env["name"]: env for env in environments}

        if len(environments) == 1:
            uid = environments[0]["uid"]
            update_devkit_config(environment_id=uid)
            return uid

        env_name = prompt_choices(
            "Available environments",
            "Select an environment",
            choices=envs_by_name.keys(),
        )
        return envs_by_name[env_name]["uid"]

    @cached_property
    def graph_id(self) -> str:
        return get_graph_by_slug(self.organization_id, self.graph_name)["uid"]

    @cached_property
    def graph_version_id(self):
        if self.explicit_graph_version_id:
            return self.explicit_graph_version_id
        return get_latest_graph_version(self.graph_id)["uid"]

    @cached_property
    def node_id(self) -> str:
        graph = self.graph_file_path
        node = self.node_file_path
        if not graph or not node:
            raise ValueError("Must specify a node")
        err_msg = f"Node {node} is not part of the graph at {graph}"

        try:
            node_path = node.absolute().relative_to(graph.parent)
        except Exception as e:
            raise Exception(err_msg) from e
        cfg = self._load_yaml(graph) or {}
        for node in cfg.get("functions", []):
            if node.get("node_file") == node_path.as_posix():
                id = node.get("id")
                if id:
                    return id
                raise Exception("Node does not have an id. Run [code]patterns upload")
        raise Exception(err_msg)

    @cached_property
    def graph_file_path(self) -> Path:
        if self.explicit_graph_path:
            return resolve_graph_path(self.explicit_graph_path, exists=True)
        if self.node_file_path:
            return _find_graph_file(
                self.node_file_path.parent,
                prompt=False,
                nearest=self.find_nearest_graph,
            )
        return _find_graph_file(path=None, prompt=True, nearest=self.find_nearest_graph)

    @cached_property
    def graph_directory(self) -> Path:
        return self.graph_file_path.parent

    @cached_property
    def root_graph_file(self) -> Path:
        return _find_graph_file(self.graph_directory, nearest=False)

    @cached_property
    def cfg(self) -> CliConfig:
        if self.ignore_local_cfg:
            return CliConfig()
        return read_devkit_config()

    @cached_property
    def graph_name(self) -> str:
        def from_yaml():
            graph = self._load_yaml(self.root_graph_file)
            return graph.get("name", self.root_graph_file.parent.name)

        if self.explicit_graph_name:
            return self.explicit_graph_name
        if self.explicit_graph_path or self.node_file_path:
            return from_yaml()
        if self.explicit_graph_version_id:
            vid = self.explicit_graph_version_id
            return get_graph_version_by_id(vid)["graph"]["name"]
        return from_yaml()

    def _load_yaml(self, path: Path) -> dict:
        with open(path) as f:
            return yaml.load(f.read(), Loader=Loader)


def _find_graph_file(
    path: Optional[Path], prompt: bool = True, nearest: bool = False
) -> Path:
    """Walk up a directory tree looking for a graph

    :param path: The location to start the search
    :param prompt: If True, ask the user to enter a path if it can't be found
    :param nearest: If False, keep walking up until there's no graph.yml in the parent
                    directory. If True, stop as soon as one if found.
    """
    if path and path.is_file():
        return resolve_graph_path(path, exists=True)
    if not path:
        path = Path(os.getcwd())
    path = path.absolute()

    found = None

    for _ in range(100):
        if not path or path == path.parent:
            break
        p = path / "graph.yml"
        if p.is_file():
            found = p
            if nearest:
                break
        elif found:
            break
        path = path.parent
    if found:
        return found

    if prompt:
        resp = prompt_path("Enter the path to the graph yaml file", exists=True)
        return resolve_graph_path(resp, exists=True)
    else:
        raise ValueError(f"Cannot find graph.yml{f' at {path}' if path else ''}")



import os
from pathlib import Path
from typing import Optional

import platformdirs
import pydantic

DEVKIT_CONFIG_ENV_VAR = "PATTERNS_CONFIG"
DEVKIT_CONFIG_NAME = "config.json"


class AuthServer(pydantic.BaseModel):
    domain: str
    audience: str
    devkit_client_id: str


class CliConfig(pydantic.BaseModel):
    organization_id: str = None
    environment_id: str = None
    token: str = None
    refresh: str = None
    auth_server: AuthServer = None

    class Config:
        extra = "ignore"
        allow_population_by_field_name = True


def get_devkit_config_path() -> Path:
    path = os.environ.get(DEVKIT_CONFIG_ENV_VAR)
    if path:
        return Path(path)
    config_dir = platformdirs.user_config_dir("patterns", appauthor=False, roaming=True)
    return Path(config_dir) / DEVKIT_CONFIG_NAME


def read_devkit_config() -> CliConfig:
    path = get_devkit_config_path()
    if path.exists():
        return CliConfig.parse_file(path)
    return CliConfig()


def write_devkit_config(config: CliConfig):
    path = get_devkit_config_path()
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(config.json(indent="  "))


_UNCHANGED = object()


def update_devkit_config(
    organization_id: Optional[str] = _UNCHANGED,
    environment_id: Optional[str] = _UNCHANGED,
    token: Optional[str] = _UNCHANGED,
    refresh: Optional[str] = _UNCHANGED,
    auth_server: Optional[AuthServer] = _UNCHANGED,
) -> CliConfig:
    cfg = read_devkit_config()
    update = {}
    if organization_id != _UNCHANGED:
        update["organization_id"] = organization_id
    if environment_id != _UNCHANGED:
        update["environment_id"] = environment_id
    if token != _UNCHANGED:
        update["token"] = token
    if refresh != _UNCHANGED:
        update["refresh"] = refresh
    if auth_server != _UNCHANGED:
        update["auth_server"] = auth_server
    copy = cfg.copy(update=update)
    write_devkit_config(copy)
    return copy



import functools
import itertools
from typing import Callable, Iterable, List

from patterns.cli.services.api import get_json


class PaginatedCall:
    def __init__(self, initial_request: Callable[[], dict]):
        self._initial_request = initial_request

    def pages(self) -> Iterable[List[dict]]:
        """Iterate over pages returned from the endpoint"""
        data = self._initial_request()
        yield data["results"]
        while data["next"]:
            data = get_json(data["next"], base_url="")
            if data["results"]:
                yield data["results"]

    def __iter__(self) -> Iterable[dict]:
        """Iterate over all objects returned from the endpoint"""
        return itertools.chain.from_iterable(self.pages())


def paginated(fn: Callable[..., dict]) -> Callable[..., PaginatedCall]:
    """Decorator that yields pages from a paginated endpoint"""

    def f(*args, **kwargs):
        return PaginatedCall(functools.partial(fn, *args, **kwargs))

    return f



from __future__ import annotations

from patterns.cli.services.api import Endpoints, post_for_json


def create_graph_component(graph_version_uid: str,) -> dict:
    body = {"graph_version_uid": graph_version_uid}
    return post_for_json(Endpoints.COMPONENTS_CREATE, json=body)



from __future__ import annotations

from patterns.cli.services.api import Endpoints, get_json


def get_graph_by_slug(organization_uid: str, slug: str) -> dict:
    return get_json(Endpoints.graph_by_slug(organization_uid, slug))


def get_graph_version_by_id(graph_version_uid) -> dict:
    return get_json(Endpoints.graph_version_by_id(graph_version_uid))


def get_latest_graph_version(graph_uid: str) -> dict:
    return get_json(Endpoints.graphs_latest(graph_uid))["latest_graph_version"]



from __future__ import annotations

from patterns.cli.services.api import Endpoints, get_json
from patterns.cli.services.pagination import paginated


def get_organization_by_name(name: str) -> dict:
    return get_json(Endpoints.organization_by_slug(name))


def get_organization_by_id(organization_uid: str) -> dict:
    return get_json(Endpoints.organization_by_id(organization_uid))


@paginated
def paginated_organizations():
    return get_json(Endpoints.ORGANIZATIONS_LIST)



import io
import re
from pathlib import Path
from zipfile import ZipFile

import typer
from typer import Option, Argument

from patterns.cli.services.lookup import IdLookup
from patterns.cli.services.output import sprint, abort_on_error, abort
from patterns.cli.services.pull import (
    download_graph_zip,
    download_component_zip,
    COMPONENT_RE,
)
from patterns.configuration.edit import GraphDirectoryEditor, FileOverwriteError

_graph_help = "The name of a graph in your Patterns organization [default: directory name]"
_graph_version_id_help = (
    "The id of the graph version to pull. [default: latest version]"
)
_organization_help = "The name of the Patterns organization that the graph was uploaded to"
_force_help = "Overwrite existing files without prompting"
_directory_help = "The directory to create the new graph in. Must not exist."
_component_help = (
    "The component version to download (e.g. 'organization/component@v1')."
)


def clone(
    organization: str = Option("", "-o", "--organization", help=_organization_help),
    graph: str = Option("", help=_graph_help),
    graph_version_id: str = Option("", "-v", "--version", help=_graph_version_id_help),
    component: str = Option("", "--component", help=_component_help),
    directory: Path = Argument(None, exists=False, help=_graph_help),
):
    """Download the code for a graph"""
    if not graph and not directory and not component:
        if graph_version_id:
            abort(
                f"Missing graph directory argument."
                f"\ntry [code]patterns clone -v {graph_version_id} new_graph"
            )
        else:
            abort(
                f"Missing graph argument." f"\ntry [code]patterns clone graph-to-clone"
            )
    component_match = COMPONENT_RE.fullmatch(component)
    if component and not component_match:
        abort(
            "Invalid component version. Must be in the form organization/component@v1"
        )

    component_name = component_match.group(2) if component_match else None

    ids = IdLookup(
        organization_name=organization,
        explicit_graph_name=graph or component_name or directory.name,
        explicit_graph_version_id=graph_version_id,
    )
    if not directory:
        if component:
            directory = Path(component_name)
        elif graph:
            directory = Path(graph)
        elif graph_version_id:
            with abort_on_error("Error"):
                directory = Path(ids.graph_name)
        else:
            abort("Specify --graph, --graph-version-id, or a directory")

    with abort_on_error("Error cloning graph"):
        if component:
            content = download_component_zip(component)
        else:
            content = download_graph_zip(ids.graph_version_id)
        editor = GraphDirectoryEditor(directory, overwrite=False)
        with ZipFile(io.BytesIO(content), "r") as zf:
            editor.add_node_from_zip("graph.yml", "graph.yml", zf)

    sprint(f"[success]Cloned graph into {directory}")


_pull_graph_help = "The location of the graph to pull into [default: current directory]"


def pull(
    organization: str = Option("", "-o", "--organization", help=_organization_help),
    graph_version_id: str = Option("", help=_graph_version_id_help),
    force: bool = Option(False, "-f", "--force", help=_force_help),
    graph: Path = Argument(None, exists=True, help=_pull_graph_help),
):
    """Update the code for the current graph"""
    ids = IdLookup(
        organization_name=organization,
        explicit_graph_version_id=graph_version_id,
        explicit_graph_path=graph,
    )
    with abort_on_error("Error downloading graph"):
        b = io.BytesIO(download_graph_zip(ids.graph_version_id))
        editor = GraphDirectoryEditor(ids.graph_file_path, overwrite=force)

    with abort_on_error("Error downloading graph"):
        try:
            with ZipFile(b, "r") as zf:
                editor.add_node_from_zip("graph.yml", "graph.yml", zf)
        except FileOverwriteError as e:
            sprint(f"[error]{e}")
            sprint("[info]Run this command with --force to overwrite local files")
            raise typer.Exit(1)
    sprint(f"[success]Pulled graph content")



from __future__ import annotations

import os
import secrets
import string
import subprocess
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Generator


def _is_git_directory(path: Path) -> bool:
    return (path / ".git").is_dir()


def _all_files_not_gitignored(path: Path) -> Generator[Path]:
    files = subprocess.check_output(
        ["git", "-C", str(path), "ls-files", "-co", "--exclude-standard"]
    ).splitlines()
    for f in files:
        yield path / Path(f.decode())


def _all_files(path: Path) -> Generator[str]:
    for dirname, dirnames, files in os.walk(path):
        if "__pycache__" in dirnames:
            dirnames.remove("__pycache__")
        for f in files:
            p = Path(dirname) / f
            if p.suffix != ".pyc" and p.name.lower() != ".ds_store":
                yield p


def compress_directory(path: Path) -> BytesIO:
    io = BytesIO()
    zipf = zipfile.ZipFile(io, "w", zipfile.ZIP_DEFLATED)
    if _is_git_directory(path):
        # Respect .gitignore
        contents = _all_files_not_gitignored(path)
    else:
        contents = _all_files(path)
    for f in contents:
        zipf.write(f, f.relative_to(path))
    zipf.close()
    io.seek(0)
    io.name = "graph_manifest.zip"
    return io


_alphabet = string.digits + string.ascii_lowercase


def random_node_id() -> str:
    return "".join(secrets.choice(_alphabet) for _ in range(8))



from pathlib import Path


def is_relative_to(self: Path, other: Path) -> bool:
    """Backport of Path.is_relative_to, which was added in python3.9"""
    try:
        self.relative_to(other)
        return True
    except ValueError:
        return False



import contextlib
import typing
from pathlib import Path

import rich.prompt
import typer
from requests import HTTPError
from rich.console import Console
from rich.theme import Theme

"""Set to True to raise exceptions from cli commands rather than printing the message"""
DEBUG = False

console = Console(
    theme=(
        Theme(
            {
                "info": "italic cyan",
                "warning": "magenta",
                "success": "green",
                "error": "red",
                "code": "bold cyan",
            }
        )
    )
)


class _PathPrompt(rich.prompt.Prompt):
    response_type = Path
    validate_error_message = "[prompt.invalid]Please enter a valid file path"


def prompt_path(
    message: str, default: typing.Union[Path, str] = None, exists: bool = None
) -> Path:
    while True:
        p = _PathPrompt.ask(message, default=default)
        if not p:
            continue
        if exists is True and not p.exists():
            sprint("[prompt.invalid]Path already exists")
        elif exists is False and p.exists():
            sprint("[prompt.invalid]Path does not exist")
        else:
            break
    return p


def prompt_str(message: str, default: str = None, password: bool = False) -> str:
    return rich.prompt.Prompt.ask(message, default=default, password=password)


def prompt_choices(
    choice_message: str,
    prompt_message: str,
    choices: typing.Iterable[str],
) -> str:
    sprint(f"[info]{choice_message}:")
    for c in choices:
        sprint(f"    [info]{c}")
    return rich.prompt.Prompt.ask(
        prompt_message,
        choices=list(choices),
        show_choices=False,
    )


def sprint(message):
    """Print styled content"""
    console.print(message)


def abort(message: str) -> typing.NoReturn:
    """Print an error message and raise an Exit exception"""
    sprint(f"[error]{message}")
    raise typer.Exit(1)


@contextlib.contextmanager
def abort_on_error(message: str, prefix=": ", suffix=""):
    """Catch any exceptions that occur and call `abort` with their message"""
    if DEBUG:
        yield
        return
    try:
        yield
    except HTTPError as e:
        try:
            details = e.response.json()["detail"]
        except Exception:
            details = e.response.text
        if not details:
            details = f"HTTP {e.response.status_code}"

        # check 403 error message for unverified email / unsetup account and display message
        # we give them the home page, since the webapp will redirect them to the proper setup page automatically
        if e.response.status_code == 403:
            if details == "unverified email":
                abort(
                    f"Please verify your email address before using Patterns - https://studio.patterns.app"
                )
            elif details == "incomplete setup":
                abort(
                    f"Please finish account setup before using Patterns - https://studio.patterns.app"
                )
        elif e.response.status_code == 401:
            abort(
                "You are not logged in to the devkit.\n"
                "[info]You can log in with [code]patterns login"
            )

        abort(f"{message}{prefix}{details}{suffix}")
    except (typer.Exit, typer.Abort) as e:
        raise e
    except KeyError as e:
        abort(f"{message}{prefix}KeyError: {e}{suffix}")
    except Exception as e:
        abort(f"{message}{prefix}{e}{suffix}")



from __future__ import annotations

from patterns.cli.services.api import Endpoints, post_for_json


def deploy_graph_version(graph_version_uid: str, environment_uid: str) -> dict:
    return post_for_json(
        Endpoints.DEPLOYMENTS_DEPLOY,
        json={
            "environment_uid": environment_uid,
            "graph_version_uid": graph_version_uid,
        },
    )



from __future__ import annotations

import urllib.parse
import webbrowser
from abc import ABCMeta, abstractmethod
from datetime import timedelta
from http.server import HTTPServer, BaseHTTPRequestHandler
from socketserver import BaseRequestHandler
from threading import Barrier, Thread, BrokenBarrierError
from typing import Tuple, Callable, Optional, List, Dict, Type
from urllib.parse import ParseResult

from patterns.cli.services.output import abort, sprint

LOCAL_OAUTH_PORT = 30420

# Long-ish timeout.  If someone is signing up with an email / password, it can take a while.
REQUEST_TIMEOUT = timedelta(minutes=1)


def execute_oauth_flow(
    url: str,
    request_handler_class: Type[BaseOAuthRequestHandler],
    on_request: Callable[[BaseOAuthRequestHandler], None] | None = None,
):
    """Execute an oauth flow, opening a web browser and handling callbacks.

    Parameters:
        url: URL to open in the browser
        request_handler_class: Class which will handle the web request made by the
                               browser (for an oauth callback)
        on_request: Optional callback which is invoked before the actual handler is run
    """

    # noinspection PyTypeChecker
    server = OAuthHttpServer(
        ("localhost", LOCAL_OAUTH_PORT), request_handler_class, on_request
    )
    webbrowser.open(url, new=1, autoraise=True)

    server_thread = ServerThread(server)
    server_thread.start()

    try:
        server.wait_for_request_started()
    except BrokenBarrierError:
        server.shutdown()
        abort("Timed out waiting for the web browser to hit callback url")

    try:
        server.wait_for_request_finished()
    except BrokenBarrierError:
        server.shutdown()
        abort("Timed out waiting for the request to finish")

    server.shutdown()
    server_thread.join(timeout=5)

    if server.error_result:
        abort(server.error_result)
    elif server.success_result:
        sprint(f"[info]{server.success_result}")
    else:
        abort("OAuth server finished without an error or success result")


class ServerThread(Thread):
    def __init__(self, server: OAuthHttpServer):
        super().__init__()
        self._server = server

    def run(self):
        self._server.serve_forever()


class OAuthHttpServer(HTTPServer):
    """Utility base class for coordinating between handlers and the CLI"""

    def __init__(
        self,
        server_address: Tuple[str, int],
        request_handler_class: Callable[..., BaseRequestHandler],
        on_request_cb: Callable[[BaseRequestHandler], None] | None,
    ):
        super().__init__(server_address, request_handler_class)
        self._on_request_cb = on_request_cb
        self.error_result = None
        self.success_result = None

        timeout_seconds = REQUEST_TIMEOUT.total_seconds()
        self._request_started_barrier = Barrier(2, timeout=timeout_seconds)
        self._request_finished_barrier = Barrier(2, timeout=timeout_seconds)

    def wait_for_request_started(self):
        self._request_started_barrier.wait()

    def wait_for_request_finished(self):
        self._request_finished_barrier.wait()

    def on_request(self, handler: BaseRequestHandler):
        if self._on_request_cb:
            self._on_request_cb(handler)

    def finish_with_error(self, error_result: str):
        self.error_result = error_result

    def finish_with_success(self, success_result: str):
        self.success_result = success_result


class BaseOAuthRequestHandler(BaseHTTPRequestHandler, metaclass=ABCMeta):
    @property
    @abstractmethod
    def handled_path(self) -> str:
        """The path which this handler will process ('/some_callback')"""
        ...

    @abstractmethod
    def handle_callback(self, parsed_url: ParseResult):
        ...

    @property
    def oauth_http_server(self) -> OAuthHttpServer:
        # noinspection PyTypeChecker
        return self.server

    def do_GET(self):
        parsed_url = urllib.parse.urlparse(self.path)

        if parsed_url.path == self.handled_path:
            self.oauth_http_server.on_request(self)
            self.oauth_http_server._request_started_barrier.wait()
            try:
                self.handle_callback(parsed_url)
            finally:
                self.oauth_http_server._request_finished_barrier.wait()
        else:
            # We don't finish_with_error here, because the browser might request
            # all kinds of things (CORS OPTIONS or favicon.ico, etc.) in addition
            # to the request we really want.  If there is a real problem here,
            # the managing thread will shut down the server
            self.send_html_response(404, f"Unhandled path: {parsed_url.path}")

    def get_single_queryparam(
        self, param_name: str, params: Dict[str, List[str]]
    ) -> Optional[str]:
        if (
            not param_name in params
            or len(params[param_name]) != 1
            or not isinstance(params[param_name][0], str)
        ):
            self.finish_with_error(
                500, f"Invalid {param_name} in response: {params.get(param_name)}"
            )
            return None
        return params[param_name][0]

    def send_html_response(self, code: int, html: str):
        self.send_response(code)
        self.send_header("Content-type", "text/html")
        self.end_headers()

        self.wfile.write(bytes(f"<html><body>{html}</body></html>", "utf-8"))

    def finish_with_error(self, status_code: int, error_result: str):
        self.send_html_response(
            status_code, "An error occurred.  See the console logs for details."
        )
        self.oauth_http_server.finish_with_error(error_result)

    def finish_with_success(self, success_result: str, success_browser_html: str):
        self.send_html_response(200, success_browser_html)
        self.oauth_http_server.finish_with_success(success_result)



from typing import List, Optional

import click
import typer
from click import Group, Context, HelpFormatter, MultiCommand, Command

from .commands.config import config
from .commands.create import create
from .commands.deploy import deploy
from .commands.list import list_command
from .commands.login import login
from .commands.logout import logout
from .commands.pull import pull, clone
from .commands.trigger import trigger
from .commands.upload import upload
from ..cli.services import output


class _Command(Group):
    def __init__(self):
        def debug_cb(ctx, p, v):
            if v:
                output.DEBUG = True

        debug_opt = click.Option(
            ["--stacktrace"], hidden=True, is_flag=True, callback=debug_cb
        )
        super().__init__(name="patterns", no_args_is_help=True, params=[debug_opt])

    def add_typer_fn(self, fn, **kw):
        if isinstance(fn, typer.Typer):
            fn._add_completion = False
            self.add_command(typer.main.get_command(fn))
        else:
            tmp = typer.Typer(add_completion=False)
            tmp.command(**kw)(fn)
            self.add_command(typer.main.get_command(tmp))

    # override help output to include nested subcommands
    def format_commands(self, ctx: Context, formatter: HelpFormatter) -> None:
        old_list = self.list_commands
        old_get = self.get_command

        self.list_commands = self._list_commands
        self.get_command = self._get_command

        super().format_commands(ctx, formatter)

        self.list_commands = old_list
        self.get_command = old_get

    def _list_commands(self, ctx: Context) -> List[str]:
        l = super().list_commands(ctx)
        for c in l:
            sub = super().get_command(ctx, c)
            if isinstance(sub, MultiCommand):
                l.extend(f"{c} {s}" for s in sub.list_commands(ctx))
        return l

    def _get_command(self, ctx: Context, cmd_name: str) -> Optional[Command]:
        parts = cmd_name.split()
        base = super().get_command(ctx, parts[0])
        if len(parts) == 1:
            return base
        assert len(parts) == 2
        assert isinstance(base, MultiCommand)
        return base.get_command(ctx, parts[1])


app = _Command()

for command in (
    config,
    create,
    deploy,
    list_command,
    login,
    logout,
    trigger,
    upload,
    pull,
    clone,
):
    app.add_typer_fn(command)


def main():
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['patterns',
 'patterns.cli',
 'patterns.cli.commands',
 'patterns.cli.services',
 'patterns.configuration',
 'patterns.helpers',
 'patterns.node']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.0,<9.0.0',
 'platformdirs>=2.4.0,<3.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'rich>=12.0.1,<13.0.0',
 'ruyaml>=0.91.0,<0.92.0',
 'typer[all]>=0.4.1,<0.5.0']

entry_points = \
{'console_scripts': ['patterns = patterns.cli.main:main']}

setup_kwargs = {
    'name': 'patterns-devkit',
    'version': '0.1.9',
    'description': 'Data pipelines from re-usable components',
    'long_description': None,
    'author': 'AJ Alt',
    'author_email': 'aj@patterns.app',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
